/*************************************************************************
 * TLV 02/14/2002                                                        *
 * Copyright (C) 2002 Travis Veldkamp                                    *
 *                                                                       *
 * This program is for programming the I-PAC keyboard encoder            *
 * For more information, see http://www.ultimarc.com                     *
 *                                                                       *
 * This is free software; you can redistribute it and/or modify          *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation; either version 2 of the License, or     *
 * at your option) any later version.                                    *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * You should have received a copy of the GNU General Public License     *
 * along with Foobar; if not, write to:                                  *
 *                   Free Software Foundation, Inc.                      *
 *                   59 Temple Place                                     *
 *                   Suite 330                                           *
 *                   Boston, MA  02111-1307                              *
 *                   USA                                                 *
 *************************************************************************/


#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <usb.h>

#define IPAC_USB_TIMEOUT	100

/* All boards are programmed with a 96 byte block of data */
#define IPAC_DATA_LEN      	96 
#define IPAC_BLOCK_LEN 		32

/* 4 macros for each board, each composed of 4 keystrokes */
#define NUM_MACRO_KEYCODES     (4*4)

/* IPAC-2 */
#define IPAC2_NUM_KEYCODES  56
#define IPAC2_PASSES        1

/* IPAC-4 - two passes of 57 keycodes each*/
#define IPAC4_NUM_KEYCODES  56
#define IPAC4_PASSES        2

/* IPAC-VE */
#define IPACVE_NUM_KEYCODES  60
#define IPACVE_PASSES        1

/* MiniPac-VE */
#define MINIPACVE_NUM_KEYCODES  64
#define MINIPACVE_PASSES        1

/* Responses */
#define RESP_GOOD               1
#define RESP_CKSUM_IPAC2	3
#define RESP_CKSUM_IPAC4	4
#define RESP_CKSUM_IPACVE	7
#define RESP_CKSUM_MINIPACVE	8

/* Map checksum response to board name */
static const char *board_name_by_resp[RESP_CKSUM_MINIPACVE+1] = { "invalid",
                                                                  "invalid",
                                                                  "invalid",
                                                                  "ipac2",  
                                                                  "ipac4", 
                                                                  "invalid",
                                                                  "invalid",
                                                                  "ipacve", 
                                                                  "minipacve"};

/* Board types */
#define NUM_BOARD_TYPES         4
#define BOARD_TYPE_IPAC2        0
#define BOARD_TYPE_IPAC4        1
#define BOARD_TYPE_IPACVE       2
#define BOARD_TYPE_MINIPACVE    3
#define BOARD_TYPE_INVALID      -1

static const char *board_name_by_type[NUM_BOARD_TYPES] = { "ipac2",
                                                           "ipac4",
                                                           "ipacve",
                                                           "minipacve" };


/* 2nd D of array is: Number valid (shifted/unshifted) codes, num passes */
#define BOARD_INFO_INDEX_UNSHIFTED_CODES 0
#define BOARD_INFO_INDEX_SHIFTED_CODES   1
#define BOARD_INFO_INDEX_PASSES          2
static const char board_info[NUM_BOARD_TYPES][3] = { {28, 28, 1},   // IPAC2
                                               {28, 28, 2},   // IPAC4
                                               {32, 28, 1},   // IPAC VE
                                               {36, 28, 1} }; // MiniPAC VE

#define CONTROL_SELECT_PAGE_2    (1)
#define CONTROL_FACTORY_DEFAULTS (1 << 1)
#define CONTROL_RAM_ONLY         (1 << 2)

/* Programming order is: shift code, unshifted codes, shifted codes, checksum, control, macros, zero pad to 96 */


// TODO: these are unused -- remove
/* Default data */
/* byte 58 is checksum, 59 is control */
static char default_ipac2[96] = { 0x10, 0xF5, 0xF4, 0x14, 0xEB, 0x29, 0xF2, 0x1A,   /*  0-7  */
                           0x11, 0x34, 0x12, 0x23, 0x22, 0x2D, 0x2B, 0x1C,   /*  8-15 */
                           0x16, 0x1B, 0x1E, 0x15, 0x2E, 0x1D, 0x36, 0x43,   /* 16-23 */
                           0x21, 0x42, 0x2A, 0x3B, 0x4B, 0x0E, 0x0D, 0x2E,   /* 24-31 */
                           0x5A, 0x00, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 32-39 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x76, 0x00,   /* 40-47 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 48-55 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 56-63 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 64-71 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 72-79 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 80-87 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }; /* 88-95 */

// TODO: these are NOT correct
/* 2nd block data, used for IPAC4.  byte 58 is checksum, 59 is control */
static char default_ipac4[96] = { 0x10, 0xF5, 0xF4, 0x14, 0xEB, 0x29, 0xF2, 0x1A,   /*  0-7  */
                           0x11, 0x34, 0x12, 0x23, 0x22, 0x2D, 0x2B, 0x1C,   /*  8-15 */
                           0x16, 0x1B, 0x1E, 0x15, 0x2E, 0x1D, 0x36, 0x43,   /* 16-23 */
                           0x21, 0x42, 0x2A, 0x3B, 0x4B, 0x0E, 0x0D, 0x2E,   /* 24-31 */
                           0x5A, 0x00, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 32-39 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x76, 0x00,   /* 40-47 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 48-55 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 56-63 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 64-71 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 72-79 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,   /* 80-87 */
                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }; /* 88-95 */

/* RequestType field -- upper bit (direction) */
#define USB_REQ_TYPE_HOST_TO_DEV 0x00
#define USB_REQ_TYPE_DEV_TO_HOST 0x80

/* RequestType field -- bits 5/6 (type) */
#define USB_REQ_TYPE_TYPE_STD    (0)
#define USB_REQ_TYPE_TYPE_CLASS  (1 << 5)
#define USB_REQ_TYPE_TYPE_VENDOR (2 << 5)

/* RequestType field -- bits 0-4 (recipient) */
#define USB_REQ_TYPE_RECIP_DEV       0
#define USB_REQ_TYPE_RECIP_INTERFACE 1
#define USB_REQ_TYPE_RECIP_ENDPOINT  2
#define USB_REQ_TYPE_RECIP_OTHER     3

#define USB_REQUEST_TYPE_HD_VENDOR_OTHER 0x43
#define USB_REQUEST_TYPE_DH_VENDOR_OTHER 0xC3


#define IPAC_REQUEST_PROGRAM_MODE 0xE9
#define IPAC_REQUEST_SEND_BLOCK   0xEB
#define IPAC_REQUEST_RESPONSE     0xEA


#define VENDOR_ID_ULTIMARC       0xD209
#define PRODUCT_ID_IPAC_USB      0x0301 

/* Number of I-PACs connected to 1 PC that we support */
/* I can't imagine it would ever be more than this... so just use an array instead of llist */
#define MAX_IPACS	5

/* */
int get_response(usb_dev_handle *ipac_hdl);

/* sends one 32 byte block */
int send_block(usb_dev_handle *ipac_hdl, char *data);

int enter_program_mode(usb_dev_handle *ipac_hdl);

int exit_program_mode(usb_dev_handle *ipac_hdl);

int find_ipacs(struct usb_device *dev_arr[], int max_num);

int detect_ipac_model(int ipac_num, FILE *loghdl);

int program_ipac(int ipac_num, int boardtype, char *datablock, int gen_checksum, char control, FILE *loghdl);

char compute_checksum(char *datablock, int len);

usb_dev_handle *init_ipac(int ipac_num, FILE *loghdl);
int close_ipac(usb_dev_handle *ipac_hdl);

int re_enumerate(int ipac_num, FILE *loghdl);

int ipac_count(void);
